@@ -2,77 +2,83 @@
 Collection of functions related to dot-calling
 
 The main user-facing API function is:
-dots(
-    clr,
-    expected,
-    expected_value_col="balanced.avg",
-    clr_weight_name="weight",
-    view_df=None,
-    kernels=None,
-    max_loci_separation=10_000_000,
-    max_nans_tolerated=1,
-    n_lambda_bins=40,
-    lambda_bin_fdr=0.1,
-    clustering_radius=20_000,
-    cluster_filtering=None,
-    tile_size=5_000_000,
-    nproc=1,
-)
+
+.. code-block:: python
+
+    dots(
+        clr,
+        expected,
+        expected_value_col="balanced.avg",
+        clr_weight_name="weight",
+        view_df=None,
+        kernels=None,
+        max_loci_separation=10_000_000,
+        max_nans_tolerated=1,
+        n_lambda_bins=40,
+        lambda_bin_fdr=0.1,
+        clustering_radius=20_000,
+        cluster_filtering=None,
+        tile_size=5_000_000,
+        nproc=1,
+    )
+
 This function implements HiCCUPS-style dot calling, but enables user-specified
 modifications at multiple steps. The current implementation makes two passes
 over the input data, first to create a histogram of pixel enrichment values, 
 and second to extract significantly enriched pixels.
 
- * The function starts with compatibility verifications
-    for the `clr`, `expected` and `view` of interest.
- * Recommendation or verification for `kernels` is done next.
-     Custom kernels must satisfy properties including: square shape,
-     equal sizes, odd sizes, zeros in the middle, etc. By default,
-     HiCCUPS-style kernels are recommended based on the binsize.
- * Lambda bins are defined for multiple hypothesis
-     testing separately for different value ranges of the locally adjusted expected.
-     Currently, log-binned lambda-bins are hardcoded using a pre-defined
-     BASE of 2^(1/3). `n_lambda_bins` controls the total number of bins.
- * Genomic regions in the specified `view`(all chromosomes by default)
-     are split into smaller tiles of size `tile_size`.
- * `scoring_and_histogramming_step()` is performed independently
-     on the genomic tiles. In this step, locally adjusted expected is
-     calculated using convolution kernels for each pixel in the tile.
-     All surveyed pixels are histogrammed according to their adjusted 
-     expected and raw observed counts. Locally adjusted expected is 
-     not stored in memory.
- * Chunks of histograms are aggregated together and a modified BH-FDR
-     procedure is applied to the result in `determine_thresholds()`.
-     This returns thresholds for statistical significance 
-     in each lambda-bin (for observed counts), along with the adjusted
-     p-values (q-values).
- * Calculated thresholds are used to extract statistically significant
-     pixels in `scoring_and_extraction_step()`. Because locally adjusted
-     expected is not stored in memory, it is re-caluclated
-     during this step, which makes it computationally intensive.
-     Locally adjusted expected values are required in order to apply
-     different thresholds of significance depending on the lambda-bin.
- * Returned filtered pixels, or 'dots', are significantly enriched 
-     relative to their locally adjusted expecteds and thus have potential
-     biological interest. Dots are further annotated with their 
-     genomic coordinates and q-values (adjusted p-values) for
-     all applied kernels.
- * All further steps perform optional post-processing on called dots
-      - enriched pixels that are within `clustering_radius` of each other
-        are clustered together and the brightest one is selected as the
-        representative position of a dot.
-      - cluster-representatives along with "singletons" (enriched pixels
-        that are not part of any cluster) can be subjected to further
-        empirical enrichment filtering in `cluster_filtering_hiccups()`. This 
-        both requires clustered dots exceed prescribed enrichment thresholds 
-        relative to their local neighborhoods and that singletons pass an 
-        even more stringent q-value threshold.
+- The function starts with compatibility verifications   
+- Recommendation or verification for `kernels` is done next.  
+  Custom kernels must satisfy properties including: square shape,
+  equal sizes, odd sizes, zeros in the middle, etc. By default,
+  HiCCUPS-style kernels are recommended based on the binsize.
+- Lambda bins are defined for multiple hypothesis  
+  testing separately for different value ranges of the locally adjusted expected.
+  Currently, log-binned lambda-bins are hardcoded using a pre-defined
+  BASE of 2^(1/3). `n_lambda_bins` controls the total number of bins.
+  for the `clr`, `expected` and `view` of interest.
+- Genomic regions in the specified `view`(all chromosomes by default)  
+  are split into smaller tiles of size `tile_size`.
+- `scoring_and_histogramming_step()` is performed independently  
+  on the genomic tiles. In this step, locally adjusted expected is
+  calculated using convolution kernels for each pixel in the tile.
+  All surveyed pixels are histogrammed according to their adjusted 
+  expected and raw observed counts. Locally adjusted expected is 
+  not stored in memory.
+- Chunks of histograms are aggregated together and a modified BH-FDR  
+  procedure is applied to the result in `determine_thresholds()`.
+  This returns thresholds for statistical significance 
+  in each lambda-bin (for observed counts), along with the adjusted
+  p-values (q-values).
+- Calculated thresholds are used to extract statistically significant  
+  pixels in `scoring_and_extraction_step()`. Because locally adjusted
+  expected is not stored in memory, it is re-caluclated
+  during this step, which makes it computationally intensive.
+  Locally adjusted expected values are required in order to apply
+  different thresholds of significance depending on the lambda-bin.
+- Returned filtered pixels, or 'dots', are significantly enriched  
+  relative to their locally adjusted expecteds and thus have potential
+  biological interest. Dots are further annotated with their 
+  genomic coordinates and q-values (adjusted p-values) for
+  all applied kernels.
+- All further steps perform optional post-processing on called dots
+
+  - enriched pixels that are within `clustering_radius` of each other  
+    are clustered together and the brightest one is selected as the
+    representative position of a dot.
+  - cluster-representatives along with "singletons" (enriched pixels  
+    that are not part of any cluster) can be subjected to further
+    empirical enrichment filtering in `cluster_filtering_hiccups()`. This 
+    both requires clustered dots exceed prescribed enrichment thresholds 
+    relative to their local neighborhoods and that singletons pass an 
+    even more stringent q-value threshold.
 """
 
 from functools import partial, reduce
 import multiprocess as mp
 import logging
 import warnings
+import time
 
 from scipy.linalg import toeplitz
 from scipy.ndimage import convolve
@@ -93,6 +99,7 @@ from ..lib.checks import (
 
 from bioframe import make_viewframe
 
+warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
 logging.basicConfig(level=logging.INFO)
 
 # this is to mitigate and parameterize the obs.raw vs count controversy:
@@ -988,6 +995,8 @@ def determine_thresholds(gw_hist, fdr):
         # run cumulative min, on the array of adjusted p-values
         # to make sure q-values are monotonously decreasing with pvals
         qvalues[k] = qvalues[k].mask(qvalues[k] > 1.0, 1.0)
+        # cast categorical index of dtype-interval to proper interval index
+        threshold_df[k].index = pd.IntervalIndex(threshold_df[k].index)
 
     return threshold_df, qvalues
 
@@ -1020,9 +1029,12 @@ def extract_scored_pixels(scored_df, thresholds, obs_raw_name=observed_count_nam
     for kernel_name, threshold in thresholds.items():
         # locally adjusted expected (lambda) of the scored pixels:
         lambda_of_pixels = scored_df[f"la_exp.{kernel_name}.value"]
-        # extract lambda_threshold for every scored pixel based on its estimated lambda, using
-        # interval-indexed 'threshold' to query it by the lambda-values of each scored pixel
-        threshold_of_pixels = threshold.loc[lambda_of_pixels]
+        # reconstruct edges of lambda bins from threshold's index:
+        ledges_reconstruct = np.r_[threshold.index.left, threshold.index.right[-1]]
+        # find indices of lambda-bins where pixels belong
+        lbin_idx = pd.cut(lambda_of_pixels, ledges_reconstruct, labels=False)
+        # extract threholds for every pixel, based on lambda-bin each of the belongs
+        threshold_of_pixels = threshold.iloc[lbin_idx]
         compliant_pixel_masks.append(
             scored_df[obs_raw_name].to_numpy() >= threshold_of_pixels.to_numpy()
         )
@@ -1248,6 +1260,11 @@ def cluster_filtering_hiccups(
             )
         )
     )
+
+    logging.info(
+        f"filtered {enrichment_fdr_comply.sum()} out of {len(centroids)} centroids to reduce the number of false-positives"
+    )
+
     # use "enrichment_fdr_comply" to filter out non-satisfying pixels:
     return centroids[enrichment_fdr_comply].reset_index(drop=True)
 
@@ -1335,7 +1352,7 @@ def scoring_and_histogramming_step(
         # drop all lambda-bins that do not have pixels in them:
         final_hist[k] = final_hist[k].loc[:, final_hist[k].sum() > 0]
         # make sure index (observed pixels counts) is sorted
-        if not final_hist[k].index.is_monotonic:
+        if not final_hist[k].index.is_monotonic_increasing:
             raise ValueError(f"Histogram for {k}-kernel is not sorted")
     # returning filtered histogram
     return final_hist
@@ -1542,6 +1559,7 @@ def dots(
         )
     except Exception as e:
         raise ValueError("provided expected is not compatible") from e
+    expected = expected.set_index(["region1", "region2", "dist"]).sort_index()
 
     # Prepare some parameters.
     binsize = clr.binsize
@@ -1587,9 +1605,10 @@ def dots(
     )
 
     # 1. Calculate genome-wide histograms of scores.
+    time_start = time.perf_counter()
     gw_hist = scoring_and_histogramming_step(
         clr,
-        expected.set_index(["region1", "region2", "dist"]),
+        expected,
         expected_value_col=expected_value_col,
         clr_weight_name=clr_weight_name,
         tiles=tiles,
@@ -1599,16 +1618,18 @@ def dots(
         loci_separation_bins=loci_separation_bins,
         nproc=nproc,
     )
-
-    logging.info("Done building histograms ...")
+    elapsed_time = time.perf_counter() - time_start
+    logging.info(f"Done building histograms in {elapsed_time:.3f} sec ...")
 
     # 2. Determine the FDR thresholds.
     threshold_df, qvalues = determine_thresholds(gw_hist, lambda_bin_fdr)
+    logging.info("Determined thresholds for every lambda-bin ...")
 
     # 3. Filter using FDR thresholds calculated in the histogramming step
+    time_start = time.perf_counter()
     filtered_pixels = scoring_and_extraction_step(
         clr,
-        expected.set_index(["region1", "region2", "dist"]),
+        expected,
         expected_value_col=expected_value_col,
         clr_weight_name=clr_weight_name,
         tiles=tiles,
@@ -1621,6 +1642,8 @@ def dots(
         bin1_id_name="bin1_id",
         bin2_id_name="bin2_id",
     )
+    elapsed_time = time.perf_counter() - time_start
+    logging.info(f"Done extracting enriched pixels in {elapsed_time:.3f} sec ...")
 
     # 4. Post-processing
     logging.info(f"Begin post-processing of {len(filtered_pixels)} filtered pixels")
@@ -1628,7 +1651,9 @@ def dots(
 
     # annotate enriched pixels
     filtered_pixels_qvals = annotate_pixels_with_qvalues(filtered_pixels, qvalues)
-    filtered_pixels_annotated = cooler.annotate(filtered_pixels_qvals, clr.bins()[:])
+    filtered_pixels_annotated = cooler.annotate(
+        filtered_pixels_qvals, clr.bins()[["chrom", "start", "end"]], replace=True
+    )
     if not clustering_radius:
         # TODO: make sure returned DataFrame has the same columns as "postprocessed_calls"
         # columns to return before-clustering
@@ -1663,7 +1688,8 @@ def dots(
     output_cols += [f"la_exp.{k}.qval" for k in kernels]
 
     # 4b. filter by enrichment and qval
-    if cluster_filtering is None:  # default - engage HiCCUPS filtering
+    if (cluster_filtering is None) or cluster_filtering:
+        # default - engage HiCCUPS filtering
         postprocessed_calls = cluster_filtering_hiccups(centroids)
     elif not cluster_filtering:
         postprocessed_calls = centroids
